// Stack_Aligner.java

import ij.*;
import ij.gui.*;
import ij.measure.*;
import ij.plugin.filter.*;
import ij.process.*;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import astroj.*;
import java.io.File;

/**
 * Based on MultiAperture_.java
 * 
 * @author F.V. Hessman, Georg-August-Universitaet Goettingen
 * @version 1.0
 * @date 2006-Oct-10
 *
 * @version 1.1
 * @date 2006-Nov-29
 * @changes Made this version an extension of MultiAperture_ to make alignment more robust.
 *
 * @version 1.2
 * @date 2009-FEB-09
 * @changes Added whole pixel shift.
 */
public class Stack_Aligner extends MultiAperture_
	{
	boolean normalize = false;
	boolean whole = true;
    boolean firstImage = true;
    boolean startingRepositionSetting = false;
    boolean startingUseVarApSetting = false;
    boolean isVirtual = false;
    String label = "Aligned_";
    String imageFilename = "";
    String imageDirname = "";
	double[] xRef = null;
	double[] yRef = null;
	double firstImageFlux = 0.0;
    String slash = IJ.isWindows() ? "\\": "/";

	/**
	 * Standard ImageJ PluginFilter setup routine which also determines the default aperture radius.
	 */
	public int setup (String arg, ImagePlus img)
		{
		if (img == null) return DONE;		// ONLY WORKS IF THERE'S AN IMAGE
		IJ.register(Stack_Aligner.class);
        startingRepositionSetting = Prefs.get (AP_PREFS_REPOSITION, true);
        startingUseVarApSetting = Prefs.get(MultiAperture_.PREFS_USEVARSIZEAP, false);
        Prefs.set (MultiAperture_.PREFS_USEVARSIZEAP, false);
        Prefs.set (AP_PREFS_REPOSITION, true);
        imageDirname = img.getOriginalFileInfo().directory+"aligned"+slash;
        isVirtual = img.getStack().isVirtual();
        if (isVirtual)
            {
            File dir = new File(imageDirname);
            if (!dir.exists())
                {
                dir.mkdir();
                }
            else if (dir.isFile())
                {
                IJ.beep();
                IJ.showMessage("Aligning with a virtual stack requires the sub-directory name 'aligned' to be available.\n"+
                               "A file named 'aligned' in the stack directory is blocking the creation of the sub-directory.");
                return DONE;
                }
            }
		return super.setup(arg,img);
		}

	/**
	 * Initializes the reference position arrays.
	 */
	protected boolean prepare ()
		{
		if (stackSize < 2) return false;
		doStack = true;
		return super.prepare();
		}

	/**
	 * Adds the aperture parameters to the list of apertures.
	 */
	protected void addAperture ()
		{
		super.addAperture ();
		}

	/**
	 * Dialog for this MultiAperture_ sub-class
	 */
	protected GenericDialog dialog()
		{
		// CREATE DIALOGUE WINDOW
		GenericDialog gd = new GenericDialog("Stack Aligner");

		// REQUIRED FIELDS

		gd.addNumericField ("   Maximum number of alignment objects per image :", nAperturesMax,0,6,"  (right click in image to finalize)");
        gd.addMessage("");
		if (stackSize > 1)
			{
            gd.addSlider("           First slice ", 1, stackSize, firstSlice == stackSize ? 1 : firstSlice);
            gd.addSlider("           Last slice ", 1, stackSize, lastSlice);
//			gd.addNumericField ("           First slice :", firstSlice == stackSize ? 1 : firstSlice,0);
//			gd.addNumericField ("           Last slice :",  lastSlice, 0);
	        }
        gd.addMessage("");
        gd.addSlider("Radius of object aperture", 1, radius>100?radius:100, radius);
        gd.addSlider("Inner radius of background annulus", 1, rBack1>100?rBack1:100, rBack1);
        gd.addSlider("Outer radius of background annulus", 1, rBack2>100?rBack2:100, rBack2);
        gd.addMessage("");

		gd.addCheckbox ("Use previous "+nAperturesStored+" apertures (1-click for first aperture).",previous && nAperturesStored > 0);
		gd.addCheckbox ("Use single step mode (set new aperture locations for each image - right click to exit).",singleStep);
		gd.addMessage (" ");
//		gd.addCheckbox ("Put results in stack's own measurements table.", !oneTable);
//		gd.addCheckbox ("All measurements from one image on the same line.",wideTable);
		gd.addMessage (" ");

		// NON-REQUIRED FIELDS (mirrored in finishFancyDialog())

        normalize = Prefs.get ("stackAligner.normalize", normalize);
        whole = Prefs.get ("stackAligner.whole", whole);
		gd.addCheckbox ("Remove background and scale to common level", normalize);
		gd.addCheckbox ("Align only to whole pixels (no interpolation)!",whole);
        if (imp.getStack().isVirtual())
            {
            gd.addMessage ("NOTE: ***THIS IS A VIRTUAL STACK***\nAligned images will be placed in the sub-directory 'aligned'.\n"+ 
                           "The new aligned stack must be opened after alignment processing is finished.\n");

            }
		gd.addMessage ("Press \"OK\" and select image alignment object(s) with left clicks.\nThen right click to begin alignment process.\n"+
		               "To abort alignment object selection or processing, press <ESC>.");
		return gd;
		}

	/**
	 * Parses the non-required fields of the dialog and cleans up thereafter.
	 */
	protected boolean finishFancyDialog (GenericDialog gd)
		{
		normalize = gd.getNextBoolean();
		whole     = gd.getNextBoolean();

        Prefs.set ("stackAligner.normalize", normalize);
        Prefs.set ("stackAligner.whole", whole);   

		xPos = new double[nApertures];
		yPos = new double[nApertures];
		ngot = 0;
		xRef = new double[nApertures];
		yRef = new double[nApertures];
		return true;
		}

	/**
	 * Perform photometry on each image of selected sub-stack.
	 */
	protected void processStack ()
		{
		// GET MEAN APERTURE BRIGHTNESS

        if (firstImage)
            {
            for (int i=0; i < nApertures; i++)
                {
                xRef[i] = xPos[i];
                yRef[i] = yPos[i];
                }
            String titl = imp.getShortTitle();
            if (titl == null)
                titl = imp.getTitle();
            label += titl;
            }

		// PROCESS STACK

		super.processStack ();

		// RENAME RESULTING ALIGNED STACK

		imp.setTitle(label);
        Prefs.set (AP_PREFS_REPOSITION, startingRepositionSetting);
        Prefs.set (MultiAperture_.PREFS_USEVARSIZEAP, startingUseVarApSetting);
        if (isVirtual) IJ.showMessage("Virtual stack alignment complete.\nAligned images are saved in subdirectory 'aligned'."); 
		}

	/**
	 * Performs processing of single images.
	 */
	protected void processImage ()
		{
		// NORMAL APERTURE MEASUREMENTS (INCLUDING MEAN back and source)

		super.processImage ();

        if (firstImage)
            firstImageFlux = (src[0]+tot[0])/nApertures;

		// PERFORM BACKGROUND SUBTRACTION AND NORMALIZATION

		if (normalize)
			{
			ip.resetRoi();		// SO CAN PERFORM ON WHOLE IMAGE
			ip.add (-back);	// REMOVE MEAN BACKGROUND
			if (!firstImage)		// NORMALIZE TO STANDARD FLUX
				ip.multiply (firstImageFlux/((src[0]+tot[0])/nApertures));
			}

		// SHIFT IMAGE

		double dx = 0.0;
		double dy = 0.0;
		for (int i=0; i < nApertures; i++)
			{
			dx += xPos[i]-xRef[i];
			dy += yPos[i]-yRef[i];
			}
		dx /= nApertures;
		dy /= nApertures;
        
        if (isVirtual)
            {
            imageFilename = imp.getStack().getShortSliceLabel(slice);
            ImagePlus imp2 = (ImagePlus)imp.clone();
            imp2.setProcessor ("Aligned_"+imp.getStack().getShortSliceLabel(slice), shiftedImage(dx,dy));
            String[] scienceHeader = FitsJ.getHeader(imp2);
            if (scienceHeader != null)
                {
                if (whole)
                    {
                    scienceHeader = FitsJ.setCard("X_SHIFT", -(int)(Math.floor(dx+Centroid.PIXELCENTER)), "AIJ stack align X-shift value", scienceHeader);
                    scienceHeader = FitsJ.setCard("Y_SHIFT", -(int)(Math.floor(dy+Centroid.PIXELCENTER)), "AIJ stack align Y-shift value", scienceHeader);
                    }
                else
                    {

                    scienceHeader = FitsJ.setCard("X_SHIFT", -dx, "AIJ stack align X-shift value", scienceHeader);
                    scienceHeader = FitsJ.setCard("Y_SHIFT", -dy, "AIJ stack align Y-shift value", scienceHeader);               
                    }
                FitsJ.putHeader(imp2, scienceHeader);
                }
            IJ.saveAs(imp2,""+imageFilename.substring(imageFilename.lastIndexOf('.')),""+imageDirname+slash+"aligned_"+imageFilename);
            }
        else
            {
            imp.setProcessor ("Aligned_"+imp.getStack().getShortSliceLabel(slice), shiftedImage(dx,dy));
            String[] scienceHeader = FitsJ.getHeader(imp);
            if (scienceHeader != null)
                {
                if (whole)
                    {
                    scienceHeader = FitsJ.setCard("X_SHIFT", -(int)(Math.floor(dx+Centroid.PIXELCENTER)), "AIJ stack align X-shift value", scienceHeader);
                    scienceHeader = FitsJ.setCard("Y_SHIFT", -(int)(Math.floor(dy+Centroid.PIXELCENTER)), "AIJ stack align Y-shift value", scienceHeader);
                    }
                else
                    {

                    scienceHeader = FitsJ.setCard("X_SHIFT", -dx, "AIJ stack align X-shift value", scienceHeader);
                    scienceHeader = FitsJ.setCard("Y_SHIFT", -dy, "AIJ stack align Y-shift value", scienceHeader);               
                    }
                FitsJ.putHeader(imp, scienceHeader);  
                }
            }
        firstImage = false;
		}

	/**
	 * Shifts image linearly by an amount (dx,dy).
	 */
	protected ImageProcessor shiftedImage (double dx, double dy)
		{
		int h = ip.getHeight();
		int w = ip.getWidth();
        ImageProcessor ips = ip.duplicate();

		for (int j=0; j < h; j++)
			{
			double y = (double)j+dy;
			int iy = (int)(y+Centroid.PIXELCENTER);
			if (whole)
				{
				for (int i=0; i < w; i++)
					{
					double x = (double)i+dx;
					int ix = (int)(x+Centroid.PIXELCENTER);
					double d = (double)ip.getPixelValue(ix, iy);
					ips.putPixelValue (i,j,d);
					}
				}
			else	{
				for (int i=0; i < w; i++)
					{
					double x = (double)i+dx;
					double d = ip.getInterpolatedPixel(x, y);
					ips.putPixelValue (i,j,d);
					}
				}
			}
		return ips;
		}

	}
